#include "mainwidget.h"
#include "ui_mainwidget.h"
#include <QDateTime>
#include <QDebug>

MainWidget::MainWidget(QWidget *parent) : QWidget(parent), ui(new Ui::MainWidget)
{
  ui->setupUi(this);
  model = new QFileSystemModel(this);
  model->setFilter(QDir::QDir::AllEntries);
  model->setRootPath("");
  ui->lvBackup->setModel(model);
  ui->lvSource->setModel(model);

  //создаем наследника QThread (для автоматического удаления можно указать объект-родитель)
  //сам объект создается в основном потоке
  worker = new BackupWorker(this);

  connect(ui->lvBackup, SIGNAL(doubleClicked(QModelIndex)),
          this, SLOT(on_lvSource_doubleClicked(QModelIndex))); 

  //соединяем слот завершения потока со слотом подготовки к следующему запуску
  connect(worker, SIGNAL(finished()), this, SLOT(readyToStart()));
}

MainWidget::~MainWidget()
{
  //Запрос приостановки фонового потока
  worker->requestInterruption();
  //Ожидание завершения фонового потока 10 секунд
  worker->wait(10000);
  delete ui;
  //объект наследника QThread будет в любом случае удален как дочерний
}

//см. видео "Qt - Работа с файловой системой
void BackupWorker::contentDifference(QDir &sDir, QDir &dDir, QFileInfoList &diffList)
{
  foreach(QFileInfo sInfo, sDir.entryInfoList(QDir::Files|QDir::Dirs|QDir::NoDotAndDotDot, QDir::Name|QDir::DirsFirst)){
    bool fileExists = false;
    foreach(QFileInfo dInfo, dDir.entryInfoList(QDir::Files|QDir::Dirs|QDir::NoDotAndDotDot, QDir::Name|QDir::DirsFirst)){
      if (sInfo.fileName() == dInfo.fileName()){
        if (sInfo.isDir() || sInfo.lastModified() <= dInfo.lastModified())
          fileExists = true;
        break;
      }
    }
    if (!fileExists)
      diffList.append(sInfo);
    if (sInfo.isFile())
      continue;
    if (fileExists){
      sDir.cd(sInfo.fileName());
      dDir.cd(sInfo.fileName());
      contentDifference(sDir, dDir, diffList);
      sDir.cdUp();
      dDir.cdUp();
    }
    else {
      sDir.cd(sInfo.fileName());
      recursiveContentList(sDir, diffList);
      sDir.cdUp();
    }
  }
}

//см. видео "Qt - Работа с файловой системой
void BackupWorker::recursiveContentList(QDir &dir, QFileInfoList &contentList)
{
  foreach(QFileInfo info, dir.entryInfoList(QDir::Files|QDir::Dirs|QDir::NoDotAndDotDot, QDir::Name|QDir::DirsFirst)){
    contentList.append(info);
    if (info.isDir() && dir.cd(info.fileName())){
      recursiveContentList(dir, contentList);
      dir.cdUp();
    }
  }
}

//см. видео "Qt - Работа с файловой системой
void MainWidget::on_lvSource_doubleClicked(const QModelIndex &index)
{
  QListView* listView = (QListView*)sender();
  QFileInfo fileInfo = model->fileInfo(index);
  if (fileInfo.fileName() == ".."){
    QDir dir = fileInfo.dir();
    dir.cdUp();
    listView->setRootIndex(model->index(dir.absolutePath()));
  }
  else if (fileInfo.fileName() == "."){
    listView->setRootIndex(model->index(""));
  }
  else if (fileInfo.isDir()){
    listView->setRootIndex(index);
  }
}

void MainWidget::on_btnBackup_clicked()
{
  //Получение актуальных путей директории источника и резервного копирования
  QString sDir = model->filePath(ui->lvSource->rootIndex());
  QString dDir = model->filePath(ui->lvBackup->rootIndex());

  //Если поток запущен, необходимо завершить работу фонового потока
  if (worker->isRunning()) {
    //запрос на завершение фонового потока
    worker->requestInterruption();
    //блокируем кнопку, пока поток реально не завершится
    ui->btnBackup->setEnabled(false);
  }
  else {
    //инициализируем рабочий объект актуальными значениями путей
    worker->sDir = sDir;
    worker->dDir = dDir;
    //запускаем поток фоновой задачи
    worker->start();
    //теперь нажатие на кнопку должно останавливать фоновую задачу
    ui->btnBackup->setText("Stop");
  }
}

void MainWidget::readyToStart()
{
  //когда фоновый поток завершен
  //нажатие на кнопку должно запускать фоновую задачу
  ui->btnBackup->setText("Start");
  //и кнопку можно уже активировать
  ui->btnBackup->setEnabled(true);
}


BackupWorker::BackupWorker(QObject *parent) : QThread(parent)
{
}

BackupWorker::~BackupWorker()
{  
  qDebug() << "worked destroyed";
}

//см. видео "Qt - Работа с файловой системой
void BackupWorker::runBackup(QString sPath, QString dPath)
{  
  QDir sDir = QDir(sPath);
  QDir dDir = QDir(dPath);

  QFileInfoList diffList = QFileInfoList();
  contentDifference(sDir, dDir, diffList);

  foreach(QFileInfo diffInfo, diffList){
    QString backupPath = diffInfo.filePath().replace(sDir.absolutePath(), dDir.absolutePath());
    if (diffInfo.isFile()){
      QFile::remove(backupPath);
      QFile::copy(diffInfo.absoluteFilePath(), backupPath);
    }
    if (diffInfo.isDir()){
      dDir.mkdir(backupPath);
    }
  }
  emit backupFinished();
}


void BackupWorker::run()
{
  qDebug() << "thread started";
  forever {
    qDebug() << "task started";
    runBackup(sDir, dDir);
    //если был запрос на завершение,
    //выходим из цикла/метода и поток завершается
    if (isInterruptionRequested()) {
      break;
    }
    qDebug() << "task finished";
    qDebug() << "thread paused for 5 secs";
    sleep(5);
    qDebug() << "thread resumed";
  }
  qDebug() << "thread finished";
}
